/*
 *  Copyright (C) 1999/2000 Hendrik Lipka.
 *  All rights reserved.
 *  This program is free software; you can redistribute it and/or modify it under the terms
 *  of the GNU General Public License as published by the Free Software Foundation;
 *  either version 2 of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *  You should have received a copy of the GNU General Public License along with this
 *  program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 *  Cambridge, MA 02139, USA or see http://www.gnu.org
 *  If you encounter problems or have a suggestion, please mail to:
 *  q3master@hendriklipka.de
 */
package de.hendriklipka.q3master;

import java.net.*;

import java.util.*;

import org.apache.log4j.*;
import de.hendriklipka.q3master.handler.heartbeat.QuakeArenaHeartbeat;
import de.hendriklipka.q3master.handler.heartbeat.FlatlineHeartbeat;

/**
 *  Description of the Class
 *
 *@author     Administrator
 *@created    24. November 2001
 */
public class RequestHandler extends Thread
{

    // we can use instance variables, because we create a new class instance per request
    /**
     *  Description of the Field
     */
    protected byte[] _request = null;
    /**
     *  Description of the Field
     */
    protected InetAddress _address = null;
    /**
     *  Description of the Field
     */
    protected int _port = -1;
    /**
     *  Description of the Field
     */
    protected DatagramSocket _socket = null;
    /**
     *  Description of the Field
     */
    protected ServerStateManager _handler = null;
    
    private static Category CAT = Category.getInstance(RequestHandler.class.getName());
    private static Category LOG_CAT = Category.getInstance("q3master.request");

    /**
     *  Description of the Field
     */
    protected final static String REQ_STR_HEARTBEAT = "heartbeat ";
//    protected final static String REQ_STR_HEARTBEAT = "heartbeat QuakeArena";
    /**
     *  Description of the Field
     */
    protected final static String REQ_STR_GETSERVERS = "getservers";
    /**
     *  Description of the Field
     */
    protected final static String REQ_STR_REGISTRATION = "register";
    
    protected static Map _heartbeatHandlers=new HashMap();
    
    static
    {
        _heartbeatHandlers.put("QuakeArena-1",new QuakeArenaHeartbeat());
        _heartbeatHandlers.put("Wolfenstein-1",new QuakeArenaHeartbeat());
        _heartbeatHandlers.put("WolfFlatline-1",new FlatlineHeartbeat());
        _heartbeatHandlers.put("EnemyTerritory-1",new QuakeArenaHeartbeat());
        _heartbeatHandlers.put("ETFlatline-1",new FlatlineHeartbeat());
    }


    /**
     *  Constructor for the RequestHandler object
     *
     *@param  requestData  Description of Parameter
     *@param  address      Description of Parameter
     *@param  port         Description of Parameter
     *@param  socket       Description of Parameter
     *@param  handler      Description of Parameter
     */
    public RequestHandler(byte[] requestData, InetAddress address, int port, DatagramSocket socket, ServerStateManager handler)
    {
        super("RequestHandler_"+address+":"+port);
        _handler = handler;
        _request = requestData;
        _address = address;
        _port = port;
        _socket = socket;
    }

    public static void configure(Configuration config)
    {
        Map handlers=config.getHeartbeatHandlers();
        Iterator it=handlers.keySet().iterator();
        while (it.hasNext())
        {
            String type=(String)it.next();
            String className=(String)handlers.get(type);
            try
            {
                Class cl=Class.forName(className);
                HeartbeatHandler hbh=(HeartbeatHandler)cl.newInstance();
                _heartbeatHandlers.put(type,hbh);
            }
            catch (Exception e)
            {
                CAT.warn("cannot instantiate heartbeat handler class '"+className+"' for type "+type);
            }
        }
    }

    /**
     *  Main processing method for the RequestHandler object
     */
    public void run()
    {
        /*
         *  dumping request
         *  for (int i=0;i<_request.length;i++)
         *  {
         *  System.out.println("#"+i+":"+_request[i]+"="+(char)_request[i]);
         *  }
         */
        // request should be at least 16 bytes
        if (_request.length < 16)
        {
            LOG_CAT.error("request too short (" + _request.length + " bytes)");
            return;
        }
        // requests start with 4 bytes 0xFF
        for (int i = 0; i < 4; i++)
        {
            if (((byte)0xff) != _request[i])
            {
                LOG_CAT.error("byte #" + i + " should be 0xff, but it is " + _request[i]);
                return;
            }
        }
        // check if heartbeat
        if (_request.length > (4 + REQ_STR_HEARTBEAT.length()))
        {
            boolean found = true;
            for (int i = 4; i < (4 + REQ_STR_HEARTBEAT.length()); i++)
            {
                byte b = _request[i];
                char c = REQ_STR_HEARTBEAT.charAt(i - 4);
                if (b != c)
                {
                    found = false;
                    break;
                }
            }
            if (found)
            {
//                System.out.println("server type="+(char)_request[_request.length-2]);
                LOG_CAT.info("got a heartbeat from " + _address.getHostAddress()+":"+_port);
                String heartbeatType=new String(_request,REQ_STR_HEARTBEAT.length()+4,_request.length-(REQ_STR_HEARTBEAT.length()+4));
                heartbeatType=heartbeatType.trim();
                CAT.info("heartbeat type from " + _address.getHostAddress()+" is "+heartbeatType);
                handleHeartBeat(heartbeatType);
                return;
            }
        }

        // check if getserver request
        if (_request.length > (4 + REQ_STR_GETSERVERS.length()))
        {
            boolean found = true;
            for (int i = 4; i < (4 + REQ_STR_GETSERVERS.length()); i++)
            {
                byte b = _request[i];
                char c = REQ_STR_GETSERVERS.charAt(i - 4);
                if (b != c)
                {
                    found = false;
                    break;
                }
            }
            if (found)
            {
                handleGetServer();
                return;
            }
        }
        // check if getserver request
        if (_request.length >= (4 + REQ_STR_REGISTRATION.length() + 6))
        {
            boolean found = true;
            for (int i = 4; i < (4 + REQ_STR_REGISTRATION.length()); i++)
            {
                byte b = _request[i];
                char c = REQ_STR_REGISTRATION.charAt(i - 4);
                if (b != c)
                {
                    found = false;
                    break;
                }
            }
            if (found)
            {
                handleRegistration();
                return;
            }
        }
        LOG_CAT.error("error: unkown request!");
    }


    /**
     */
    public void handleHeartBeat(String heartbeatType)
    {
        HeartbeatHandler handler=(HeartbeatHandler)_heartbeatHandlers.get(heartbeatType);
        if (null!=handler)
        {
            handler.handleHeartbeat(_address,_port,_handler);
        }
        else
        {
            CAT.warn("no known handler for heartbeat type "+heartbeatType);
        }
    }


    /**
     */
    public void handleGetServer()
    {
        int protocol = -1;
        boolean full = false;
        boolean empty = false;
        String s = new String(_request, REQ_STR_GETSERVERS.length() + 5, _request.length - (REQ_STR_GETSERVERS.length() + 5));
        LOG_CAT.info("got a getServers request: <" + s + "> from "+ _address.getHostAddress()+":"+_port);
        StringTokenizer st = new StringTokenizer(s, " ");
        try
        {
            String prot = (String)st.nextElement();
            protocol = Integer.parseInt(prot);
            String first = (String)st.nextElement();
            if (first.equals("empty"))
            {
                empty = true;
            }
            if (first.equals("full"))
            {
                full = true;
            }
            // max 2 specifiers for server filtering
            // @todo use unlimited options - maybe other mods use more values...
            first = (String)st.nextElement();
            if (first.equals("empty"))
            {
                empty = true;
            }
            if (first.equals("full"))
            {
                full = true;
            }
        }
        // ignore all exceptions - they can happen...
        catch (NoSuchElementException ex)
        {
        }
        catch (NumberFormatException ex)
        {
        }
        CAT.debug("protocol=" + protocol);
        if (full)
        {
            CAT.debug("include full servers");
        }
        if (empty)
        {
            CAT.debug("include empty servers");
        }
        ServerData[] sda = _handler.getAllServers(protocol, full, empty);
        if (null == sda)
        {
            CAT.debug("got no servers from handler");
            return;
        }
        CAT.info("got " + sda.length + " servers from handler");
        ProtocolVersionHandler pvh = ProtocolVersionHandlerFactory.getProtocolVersionHandler(protocol);
        if (null == pvh)
        {
            LOG_CAT.error("error: can't create ProtocolHandler for protocol version " + protocol);
            return;
        }
        byte[][] responses = pvh.createResponse(sda);
        if (null == responses)
        {
            LOG_CAT.error("error: ProtocolHandler doesn't create response :(");
            return;
        }
        sendResponses(responses, pvh);
    }


    /**
     *  Description of the Method
     */
    public void handleRegistration()
    {
        CAT.debug("got registration request");
        int[] addr = new int[4];
        int[] port = new int[2];
        int offset = 4 + REQ_STR_REGISTRATION.length();
        for (int i = 0; i < 4; i++)
        {
            // handle values > 127
            byte b = _request[i + offset];
            if (b < 0)
            {
                addr[i] = 256 + b;
            }
            else
            {
                addr[i] = b;
            }
        }
        port[0] = _request[offset + 4];
        port[1] = _request[offset + 5];
        // handle values > 127
        if (0 > port[0])
        {
            port[0] = 256 + port[0];
        }
        if (0 > port[1])
        {
            port[1] = 256 + port[1];
        }
        // create values
        int serverPort = port[0] * 256 + port[1];
        String address = "" + addr[0] + "." + addr[1] + "." + addr[2] + "." + addr[3];
        ServerData sd = new ServerData();
        try
        {
            sd._serverAddress = InetAddress.getByName(address);
        }
        catch (UnknownHostException ex)
        {
            CAT.error(ex);
        }
        sd._serverPort = serverPort;
        sd._lastHeartBeat = System.currentTimeMillis();
        // add server
        CAT.debug("registering server");
        _handler.addServerByRegistration(sd);
    }


    /**
     *  Description of the Method
     *
     *@param  responses  Description of Parameter
     *@param  pvh        Description of Parameter
     */
    protected void sendResponses(byte[][] responses, ProtocolVersionHandler pvh)
    {
        int numPerPacket=pvh.getMaxBlocksPerResponse();
        int bytesPerBlock=pvh.getBytesPerBlock();
        int blockCount=responses.length;
        int blockOffset=0;
        while (true)
        {
            int blocks=Math.min(blockCount,numPerPacket);
            byte[] hdr=pvh.createResponseHeader();
            byte[] ftr=pvh.createResponseFooter();
            byte[] packetData=new byte[hdr.length+ftr.length+blocks*bytesPerBlock];
            System.arraycopy(hdr,0,packetData,0,hdr.length);
            for (int i=0;i<blocks;i++)
            {
                System.arraycopy(responses[i+blockOffset],0,packetData,
                    hdr.length+i*bytesPerBlock,bytesPerBlock);
            }
            System.arraycopy(ftr,0,packetData,hdr.length+blocks*bytesPerBlock,ftr.length);

            try
            {
                // dump - for debugging purposes
                /*
                 *  for (int j=0;j<response.length;j++)
                 *  {
                 *  int ib=(response[j]<0)?(256+response[j]):(response[j]);
                 *  System.out.println("#"+j+":"+ib+"="+(char)response[j]);
                 *  }
                 */
                DatagramPacket packet = new DatagramPacket(packetData, packetData.length, _address, _port);
                CAT.debug("sending reponse (" + packetData.length + " bytes) to " + packet.getAddress().getHostAddress() + ":" + packet.getPort());
                Object o = new Object();
                _socket.send(packet);
                // if we send out packets too fast, clients (like GameSpy) will miss packets
                // and don't display all servers :(
                synchronized(this)
                {
                    wait(500);
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
            if (blocks<numPerPacket||(blockCount==numPerPacket))
                break;
            blockOffset+=numPerPacket;
            blockCount-=numPerPacket;
        }
    }
}
