/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
q3master@hendriklipka.de
*/

package de.hendriklipka.q3master;

import java.net.*;
import java.io.IOException;

import org.apache.log4j.*;

public class Util
{
    private static Category CAT=Category.getInstance(Util.class.getName());
    private static Category LOG_CAT=Category.getInstance("q3master.util");

    public static String getAddressFromString(String addr)
    {
        int pos=addr.indexOf(':');
        if (-1==pos)
            return addr;
        return addr.substring(0,pos);
    }

    public static int getPortFromString(String addr, int def)
    {
        int pos=addr.indexOf(':');
        if (-1==pos)
            return def;
        try
        {
            return Integer.parseInt(addr.substring(pos+1));
        }
        catch (Exception ex) {}
        return def;
    }

    /**
     * send broadcast packet
     * heavily copied from ServerStatusParser - so have a look there...
     */
    protected static void sendRequest(String address, byte[] command, DatagramSocket socket, int port) throws IOException
    {
        boolean ownSocket=false;
        if (null==socket)
        {
            ownSocket=true;
            socket=de.hendriklipka.util.UDPSocketPool.getSocket();
        }
        // create packet data
        InetAddress addr=InetAddress.getByName(Util.getAddressFromString(address));
        port=Util.getPortFromString(address,port);

        // create packet
        int cmdLen=command.length;
        byte[] ba=new byte[cmdLen+5];
        ba[0]=(byte)255;
        ba[1]=(byte)255;
        ba[2]=(byte)255;
        ba[3]=(byte)255;
        for (int i=0;i<cmdLen;i++)
        {
            ba[4+i]=command[i];
        }
        // send
        DatagramPacket dp= new DatagramPacket(ba,ba.length,addr,port);
        socket.send(dp);
        if (ownSocket)
            de.hendriklipka.util.UDPSocketPool.freeSocket(socket);
    }
}